(let ((gc-cons-threshold 30000))
  (require 'cc-mode)
  ;(require 'fortran)
  (load "sunpro-init")
  ;;(load "sunpro-menubar")
  ;;(load "sunpro-keys")
  ;;(load "sccs")
  ;;(load "sunpro-sparcworks")
  (load "eos/sun-eos-load")
  (require 'annotations)
  (when (featurep 'mule)
    (load "mime-setup"))
  (garbage-collect))
